from typing import Any, Callable, Iterable, Optional, Type, Union, cast
from pathlib import Path
import cv2 as cv
import numpy as np
from .cvtypes import Size


class VideoCaptureProperty:
    """Descriptors to alias cap.get(cv.CAP_PROP_*) and cap.set(cv.CAP_PROP_*, value).

    Raises AttributeError when setting if that property is not supported.
    """

    _set_err: str = (
        "Unable to set the property {p}. The property might not be supported by\n"
        "the backend used by the cv.VideoCapture() instance."
    )

    def __init__(self, prop: str):
        self.prop = prop

    def __get__(
        self, obj: "VideoCapture", objtype: Optional[Type["VideoCapture"]] = None
    ) -> float:
        return cast(float, obj.cap.get(self.prop))

    def __set__(self, obj: "VideoCapture", value: float) -> None:
        if not obj.cap.set(self.prop, value):
            raise AttributeError(self._set_err.format(p=self.prop))


class VideoCapture:
    """An adapter for `cv.VideoCapture`, giving a more Pythonic interface."""

    pos_msec = VideoCaptureProperty(cv.CAP_PROP_POS_MSEC)
    pos_frames = VideoCaptureProperty(cv.CAP_PROP_POS_FRAMES)
    pos_avi_ratio = VideoCaptureProperty(cv.CAP_PROP_POS_AVI_RATIO)
    frame_width = VideoCaptureProperty(cv.CAP_PROP_FRAME_WIDTH)
    frame_height = VideoCaptureProperty(cv.CAP_PROP_FRAME_HEIGHT)
    fps = VideoCaptureProperty(cv.CAP_PROP_FPS)
    fourcc = VideoCaptureProperty(cv.CAP_PROP_FOURCC)
    frame_count = VideoCaptureProperty(cv.CAP_PROP_FRAME_COUNT)
    format = VideoCaptureProperty(cv.CAP_PROP_FORMAT)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.cap = cv.VideoCapture(*args, **kwargs)
        if not self.cap.isOpened():
            raise ValueError(
                f"Unable to open video source: args: {args} kwargs: {kwargs}"
            )

    def __getattr__(self, key: str) -> Any:
        return getattr(self.cap, key)

    def __iter__(self) -> Iterable[np.ndarray]:
        """Iterate through frames in the video."""
        noread = (False, None)
        if self.cap.isOpened():
            for _, frame in iter(self.cap.read, noread):
                yield frame

    def __enter__(self) -> "VideoCapture":
        """Enter the context manager."""
        return self

    def __exit__(self, *args: Any, **kwargs: Any) -> None:
        """Releases the video capture object on exiting the context manager."""
        self.cap.release()


class VideoWriter:
    filename: str
    fourcc: int
    fps: float

    _nowriter = object()

    def __init__(
        self,
        filename: Union[Path, str],
        fourcc: Union[int, Iterable[str]] = "mp4v",
        fps: float = 30,
        frameSize: Any = None,
        **kwargs: Any,
    ) -> None:
        self.filename = str(filename)
        self.fourcc = (
            fourcc if isinstance(fourcc, int) else cv.VideoWriter_fourcc(*fourcc)
        )
        self.fps = fps
        self._kwargs = kwargs

        # wait to create writer based on first frame if size is not provided
        self._writer = (
            self._nowriter if frameSize is None else self._makewriter(frameSize)
        )

    def __enter__(self) -> "VideoWriter":
        return self

    def __exit__(self, *args: Any, **kwargs: Any) -> None:
        self.release()

    def _makewriter(self, frame_size: Any) -> cv.VideoWriter:
        return cv.VideoWriter(
            filename=self.filename,
            fourcc=self.fourcc,
            fps=self.fps,
            frameSize=frame_size,
            **self._kwargs,
        )

    def write(self, frame: np.ndarray) -> bool:
        try:
            return cast(bool, self._writer.write(frame))
        except AttributeError as e:
            if self._writer is self._nowriter:
                size = Size.from_image(frame)
                self._writer = self._makewriter(size)
                return self.write(frame)
            else:
                raise e

    def release(self) -> None:
        try:
            self._writer.release()
        except AttributeError as e:
            if self._writer is not self._nowriter:
                raise e


class VideoPlayer:
    cap: VideoCapture
    rate: float

    _actions = {"quit": {ord("\x1b"), ord("q")}}

    def __init__(self, cap: VideoCapture) -> None:
        self.cap = cap
        self.rate = int(1000 / self.cap.fps)

    def play(
        self,
        window_name: str = "VideoPlayer",
        framefunc: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        loop: bool = False,
    ) -> None:
        """Plays through the video file with OpenCV's imshow()."""
        frames = (
            self.cap.__iter__()
            if framefunc is None
            else map(framefunc, self.cap.__iter__())
        )
        for frame in frames:
            cv.imshow(window_name, frame)
            key = cv.waitKey(self.rate) & 0xFF
            if key in self._actions["quit"]:
                return

        if loop:
            return self.play(window_name, framefunc, loop)



"""
This module mimics some of the OpenCV built-in types that don't get translated
into Python directly. Generally, these just get mapped to/from tuples.

However, a lot of OpenCV code would benefit from types that describe the tuple,
in addition to named attributes. As OpenCV expects tuples for these datatypes,
subclassing from tuples and namedtuples allows for flexibility without breaking
compatibility.

Wherever it made sense, functionality was copied from OpenCV. For overloaded
CPP functions, there is not a hard rule for deciding which version becomes the
defacto Python method, but in some cases, both have been provided. Alternative
constructors are given in the usual Python way---as classmethods prepended with
the verb "from".

Some liberties have been taken with respect to naming. In particular, camelCase
method names have been translated to snake_case. Some methods and attributes
are provided that OpenCV doesn't contain. 

Aside from Scalars (where you can use a numpy array for vector operations),
the usual arithmetic operations available in OpenCV are available here.

You can average two points by adding them and dividing by two; you can add a
point to a rect to shift it; and so on.
"""

from typing import NamedTuple, Tuple, Optional, Union
from collections.abc import Sequence
import enum
import numpy as np
import cv2 as cv

SeqOperable = Union[float, Sequence, np.ndarray]


class Point(NamedTuple):
    x: float
    y: float

    def __add__(self, other: SeqOperable) -> "Point":
        return self.__class__(*(np.array(self).__add__(np.array(other))))

    def __sub__(self, other: SeqOperable) -> "Point":
        return self.__class__(*(np.array(self).__sub__(np.array(other))))

    def __mul__(self, other: SeqOperable) -> "Point":
        return self.__class__(*(np.array(self).__mul__(np.array(other))))

    def __truediv__(self, other: SeqOperable) -> "Point":
        return self.__class__(*(np.array(self).__truediv__(np.array(other))))

    def __floordiv__(self, other: SeqOperable) -> "Point":
        return self.__class__(*(np.array(self).__floordiv__(np.array(other))))

    def __rsub__(self, other: SeqOperable) -> "Point":
        return self.__class__(*(np.array(self).__rsub__(np.array(other))))

    def __rtruediv__(self, other: SeqOperable) -> "Point":
        return self.__class__(*(np.array(self).__rtruediv__(np.array(other))))

    def __rfloordiv__(self, other: SeqOperable) -> "Point":
        return self.__class__(*(np.array(self).__rfloordiv__(np.array(other))))

    def __radd__(self, other: SeqOperable) -> "Point":
        return self.__class__(*(np.array(self).__radd__(np.array(other))))

    def __rmul__(self, other: SeqOperable) -> "Point":
        return self.__class__(*(np.array(self).__rmul__(np.array(other))))

    def __pos__(self) -> "Point":
        return self.__class__(*np.array(self).__pos__())

    def __neg__(self) -> "Point":
        return self.__class__(*np.array(self).__neg__())

    def __abs__(self) -> "Point":
        return self.__class__(*np.array(self).__abs__())

    def __round__(self, ndigits: Optional[int] = None) -> "Point":
        return self.__class__(*(round(v, ndigits=ndigits) for v in self))

    def __floor__(self) -> "Point":
        return self.__class__(*np.floor(self))

    def __ceil__(self) -> "Point":
        return self.__class__(*np.ceil(self))

    def cross(self, point: "Point") -> float:
        return float(np.cross(self, point))

    def dot(self, point: "Point") -> float:
        return float(np.dot(self, point))

    def ddot(self, point: "Point") -> float:
        return self.dot(point)

    def inside(self, rect: "Rect") -> bool:
        """checks whether the point is inside the specified rectangle"""
        rect = Rect(*rect)
        return rect.contains(self)


class Point3(NamedTuple):
    x: float
    y: float
    z: float

    def __add__(self, other: SeqOperable) -> "Point3":
        return self.__class__(*(np.array(self).__add__(np.array(other))))

    def __sub__(self, other: SeqOperable) -> "Point3":
        return self.__class__(*(np.array(self).__sub__(np.array(other))))

    def __mul__(self, other: SeqOperable) -> "Point3":
        return self.__class__(*(np.array(self).__mul__(np.array(other))))

    def __truediv__(self, other: SeqOperable) -> "Point3":
        return self.__class__(*(np.array(self).__truediv__(np.array(other))))

    def __floordiv__(self, other: SeqOperable) -> "Point3":
        return self.__class__(*(np.array(self).__floordiv__(np.array(other))))

    def __rsub__(self, other: SeqOperable) -> "Point3":
        return self.__class__(*(np.array(self).__rsub__(np.array(other))))

    def __rtruediv__(self, other: SeqOperable) -> "Point3":
        return self.__class__(*(np.array(self).__rtruediv__(np.array(other))))

    def __rfloordiv__(self, other: SeqOperable) -> "Point3":
        return self.__class__(*(np.array(self).__rfloordiv__(np.array(other))))

    def __radd__(self, other: SeqOperable) -> "Point3":
        return self.__class__(*(np.array(self).__radd__(np.array(other))))

    def __rmul__(self, other: SeqOperable) -> "Point3":
        return self.__class__(*(np.array(self).__rmul__(np.array(other))))

    def __pos__(self) -> "Point3":
        return self.__class__(*np.array(self).__pos__())

    def __neg__(self) -> "Point3":
        return self.__class__(*np.array(self).__neg__())

    def __abs__(self) -> "Point3":
        return self.__class__(*np.array(self).__abs__())

    def __round__(self, ndigits: Optional[int] = None) -> "Point3":
        return self.__class__(*(round(v, ndigits=ndigits) for v in self))

    def __floor__(self) -> "Point3":
        return self.__class__(*np.floor(self))

    def __ceil__(self) -> "Point3":
        return self.__class__(*np.ceil(self))

    def cross(self, point: "Point3") -> "Point3":
        return self.__class__(*np.cross(self, point))

    def dot(self, point: "Point3") -> float:
        return float(np.dot(self, point))

    def ddot(self, point: "Point3") -> float:
        return self.dot(point)


class Size(NamedTuple):
    width: float
    height: float

    def __add__(self, other: SeqOperable) -> "Size":
        return self.__class__(*(np.array(self).__add__(np.array(other))))

    def __sub__(self, other: SeqOperable) -> "Size":
        return self.__class__(*(np.array(self).__sub__(np.array(other))))

    def __mul__(self, other: SeqOperable) -> "Size":
        return self.__class__(*(np.array(self).__mul__(np.array(other))))

    def __truediv__(self, other: SeqOperable) -> "Size":
        return self.__class__(*(np.array(self).__truediv__(np.array(other))))

    def __floordiv__(self, other: SeqOperable) -> "Size":
        return self.__class__(*(np.array(self).__floordiv__(np.array(other))))

    def __rsub__(self, other: SeqOperable) -> "Size":
        return self.__class__(*(np.array(self).__rsub__(np.array(other))))

    def __rtruediv__(self, other: SeqOperable) -> "Size":
        return self.__class__(*(np.array(self).__rtruediv__(np.array(other))))

    def __rfloordiv__(self, other: SeqOperable) -> "Size":
        return self.__class__(*(np.array(self).__rfloordiv__(np.array(other))))

    def __radd__(self, other: SeqOperable) -> "Size":
        return self.__class__(*(np.array(self).__radd__(np.array(other))))

    def __rmul__(self, other: SeqOperable) -> "Size":
        return self.__class__(*(np.array(self).__rmul__(np.array(other))))

    def __pos__(self) -> "Size":
        return self.__class__(*np.array(self).__pos__())

    def __neg__(self) -> "Size":
        return self.__class__(*np.array(self).__neg__())

    def __abs__(self) -> "Size":
        return self.__class__(*np.array(self).__abs__())

    def __round__(self, ndigits: Optional[int] = None) -> "Size":
        return self.__class__(*(round(v, ndigits=ndigits) for v in self))

    def __floor__(self) -> "Size":
        return self.__class__(*np.floor(self))

    def __ceil__(self) -> "Size":
        return self.__class__(*np.ceil(self))

    def area(self) -> float:
        return self.height * self.width

    def empty(self) -> bool:
        """true if empty"""
        return self.width <= 0 or self.height <= 0

    @classmethod
    def from_image(cls, image: np.ndarray) -> "Size":
        h, w = image.shape[:2]
        return cls(w, h)


class Rect(NamedTuple):
    """Mimics cv::Rect while maintaining compatibility with OpenCV's Python bindings.

    Reference: https://docs.opencv.org/master/d2/d44/classcv_1_1Rect__.html
    """

    x: float
    y: float
    width: float
    height: float

    def __add__(self, other: Union[Point, Size]) -> "Rect":  # type: ignore[override]
        """Shift or alter the size of the rectangle.
        rect ± point (shifting a rectangle by a certain offset)
        rect ± size (expanding or shrinking a rectangle by a certain amount)
        """
        if isinstance(other, Point):
            origin = Point(self.x + other.x, self.y + other.y)
            return self.from_origin(origin, self.size())
        elif isinstance(other, Size):
            size = Size(self.width + other.width, self.height + other.height)
            return self.from_origin(self.tl(), size)
        raise NotImplementedError(
            "Adding to a rectangle generically is ambiguous.\n"
            "Add a Point to shift the top-left point, or a Size to expand the rectangle."
        )

    def __sub__(self, other: Union[Point, Size]) -> "Rect":
        """Shift or alter the size of the rectangle.
        rect ± point (shifting a rectangle by a certain offset)
        rect ± size (expanding or shrinking a rectangle by a certain amount)
        """
        if isinstance(other, Point):
            origin = Point(self.x - other.x, self.y - other.y)
            return self.from_origin(origin, self.size())
        elif isinstance(other, Size):
            w = max(self.width - other.width, 0)
            h = max(self.height - other.height, 0)
            return self.from_origin(self.tl(), Size(w, h))
        raise NotImplementedError(
            "Subtracting from a rectangle generically is ambiguous.\n"
            "Subtract a Point to shift the top-left point, or a Size to shrink the rectangle."
        )

    def __and__(self, other: "Rect") -> "Rect":
        """rectangle intersection"""
        x = max(self.x, other.x)
        y = max(self.y, other.y)
        w = min(self.x + self.width, other.x + other.width) - x
        h = min(self.y + self.height, other.y + other.height) - y

        return (
            self.__class__(0, 0, 0, 0)
            if (w <= 0 or h <= 0)
            else self.__class__(x, y, w, h)
        )

    def __or__(self, other: "Rect") -> "Rect":
        """minimum area rectangle containing self and other."""
        if self.empty():
            return other
        elif not other.empty():
            x = min(self.x, other.x)
            y = min(self.y, other.y)
            w = max(self.x + self.width, other.x + other.width) - x
            h = max(self.y + self.height, other.y + other.height) - y
            return self.__class__(x, y, w, h)
        return self

    def tl(self) -> Point:
        """top left point"""
        return Point(self.x, self.y)

    def br(self) -> Point:
        """bottom right point"""
        return Point(self.x + self.width, self.y + self.height)

    def area(self) -> float:
        return self.height * self.width

    def size(self) -> Size:
        """size (width, height) of the rectangle"""
        return Size(self.width, self.height)

    def contains(self, point: Point) -> bool:
        """checks whether the rectangle contains the point"""
        return (
            self.x <= point.x <= self.x + self.width
            and self.y <= point.y <= self.y + self.height
        )

    def empty(self) -> bool:
        """true if empty"""
        return self.width <= 0 or self.height <= 0

    @classmethod
    def from_points(cls, top_left: Point, bottom_right: Point) -> "Rect":
        """Alternative constructor using two points."""
        x1, y1 = top_left
        x2, y2 = bottom_right
        w = x2 - x1
        h = y2 - y1
        return cls(x1, y1, w, h)

    @classmethod
    def from_origin(cls, origin: Point, size: Size) -> "Rect":
        """Alternative constructor using a point and size."""
        x, y = origin
        w, h = size
        return cls(x, y, w, h)

    @classmethod
    def from_center(cls, center: Point, size: Size) -> "Rect":
        """Alternative constructor using a center point and size."""
        w, h = size
        xc, yc = center
        x = xc - w / 2
        y = yc - h / 2
        return cls(x, y, w, h)

    def slice(self) -> Tuple[slice, slice]:
        """Returns a slice for a numpy array. Not included in OpenCV.

        img[rect.slice()] == img[rect.y : rect.y + rect.height, rect.x : rect.x + rect.width]
        """
        return slice(self.y, self.y + self.height), slice(self.x, self.x + self.width)

    def center(self) -> Point:
        """Returns the center of the rectangle as a point (xc, yc). Not included in OpenCV.

        rect.center() == (rect.x + rect.width / 2, rect.y + rect.height / 2)
        """
        return Point(self.x + self.width / 2, self.y + self.height / 2)

    def intersection(self, other: "Rect") -> float:
        """Return the area of the intersection of two rectangles. Not included in OpenCV."""
        if self.empty() or other.empty():
            return 0
        w = min(self.x + self.width, other.x + other.width) - max(self.x, other.x)
        h = min(self.y + self.height, other.y + other.height) - max(self.y, other.y)
        return w * h

    def union(self, other: "Rect") -> float:
        """Return the area of the union of two rectangles. Not included in OpenCV."""
        return self.area() + other.area() - self.intersection(other)


class RotatedRect(NamedTuple):
    center: Point
    size: Size
    angle: float

    def bounding_rect(self) -> Rect:
        """returns the minimal rectangle containing the rotated rectangle"""
        pts = self.points()
        r = Rect.from_points(
            Point(np.floor(min(pt.x for pt in pts)), np.floor(min(pt.y for pt in pts))),
            Point(np.ceil(max(pt.x for pt in pts)), np.ceil(max(pt.y for pt in pts))),
        )
        return r

    def points(self) -> Tuple[Point, Point, Point, Point]:
        """returns 4 vertices of the rectangle. The order is bottom left, top left, top right, bottom right."""
        b = np.cos(np.radians(self.angle)) * 0.5
        a = np.sin(np.radians(self.angle)) * 0.5

        pt0 = Point(
            self.center.x - a * self.size.height - b * self.size.width,
            self.center.y + b * self.size.height - a * self.size.width,
        )
        pt1 = Point(
            self.center.x + a * self.size.height - b * self.size.width,
            self.center.y - b * self.size.height - a * self.size.width,
        )

        pt2 = Point(2 * self.center.x - pt0.x, 2 * self.center.y - pt0.y)
        pt3 = Point(2 * self.center.x - pt1.x, 2 * self.center.y - pt1.y)

        return pt0, pt1, pt2, pt3

    @classmethod
    def from_points(cls, point1: Point, point2: Point, point3: Point) -> "RotatedRect":
        """Any 3 end points of the RotatedRect. They must be given in order (either clockwise or anticlockwise)."""
        center = (point1 + point3) * 0.5
        vecs = [point1 - point2, point2 - point3]
        x = max(np.linalg.norm(pt) for pt in (point1, point2, point3))
        a = min(np.linalg.norm(vecs[0]), np.linalg.norm(vecs[1]))

        # check that given sides are perpendicular
        if abs(vecs[0].dot(vecs[1])) * a > np.finfo(np.float32).eps * 9 * x * (
            np.linalg.norm(vecs[0]) * np.linalg.norm(vecs[1])
        ):
            raise ValueError(
                "The three points do not define a rotated rect. The three points should form a right triangle."
            )

        # wd_i stores which vector (0,1) or (1,2) will make the width
        # One of them will definitely have slope within -1 to 1
        wd_i = 1 if abs(vecs[1][1]) < abs(vecs[1][0]) else 0
        ht_i = (wd_i + 1) % 2

        angle = np.degrees(np.arctan2(vecs[wd_i][1], vecs[wd_i][0]))
        width = np.linalg.norm(vecs[wd_i])
        height = np.linalg.norm(vecs[ht_i])
        size = Size(width, height)

        return cls(center, size, angle)


class TermCriteria(NamedTuple):
    class Type(enum.IntFlag):  # type: ignore[misc]
        COUNT: int = cv.TermCriteria_COUNT
        MAX_ITER: int = cv.TermCriteria_MAX_ITER
        EPS: int = cv.TermCriteria_EPS

    # Without this, the MAX_ITER alias won't show up in some interpreters
    Type._member_names_ = ["COUNT", "MAX_ITER", "EPS"]  # type: ignore[misc]

    type: int = Type.COUNT
    max_count: int = 0
    epsilon: float = 0

    def is_valid(self) -> bool:
        is_count = bool(self.type & self.Type.COUNT) and self.max_count > 0
        is_eps = bool(self.type & self.Type.EPS) and not np.isnan(self.epsilon)
        return is_count or is_eps



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cvpack', 'cvpack.extras']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.5,<2.0.0', 'opencv-python!=4.2.0.32']

setup_kwargs = {
    'name': 'cvpack-alkasm',
    'version': '1.1.0',
    'description': 'Utilities for OpenCV in Python',
    'long_description': '# cvpack\n\nOpenCV extensions for more Pythonic interactions.\n\n## Install\n    \n```sh\npip install cvpack-alkasm\n```\n\n## Types\n\n`cvpack` includes types that exist in the main C++ OpenCV codebase, but that aren\'t included in the Python bindings. They are compatible as arguments to OpenCV functions, and they implement the same interfaces (with some new additions). The types that are included are `Point`, `Point3`, `Rect`, `RotatedRect`, `Size`, `TermCriteria`. They are implemented as namedtuples, and as such are immutable.\n\n```python\nimport cvpack\n\nimg = cvpack.imread("img.png")\np1 = cvpack.Point(50, 50)\np2 = cvpack.Point(100, 100)\nrect = cvpack.Rect.from_points(p1, p2)\nroi = img[rect.slice()]\nroi_size = cvpack.Size.from_image(roi)\nassert roi_size == rect.size()\n```\n\nThe overloaded constructors are available as `from_` classmethods, like `from_points` shown above. They also follow the same operator overloads that OpenCV has: two points summed is a point, adding a point to a rectangle shifts it, you can `&` two rectangles to get the intersection as a new rectangle, and so on.\n\n## Image IO\n\nWrappers for `imread`, `imwrite`, and `imshow` simplify usage by checking errors and allowing path-like objects for path arguments. Additionally, `cvpack` provides functions to read images from a URL (`imread_url`), display to a browser (`imshow_browser`) for statically serving images while working in an interpreter, and displaying images in a Jupyter notebook (`imshow_jupyter`) as HTML directly rather than the typical `plt.imshow` from `matplotlib`. Some other utilities related to display are also included.\n\n```python\nfrom pathlib import Path\nimport cvpack\n\nfor path in Path("folder").glob("*.png"):\n    img = cvpack.imread(path)\n    big = cvpack.add_grid(cvpack.enlarge(img))\n    cvpack.imshow_browser(img, route=str(path))\n```\n\n## Video IO\n\nWorking with video requires acquiring and releasing resources, so `cvpack` provides context managers for video readers and writers which wrap the classes from OpenCV. Reading video frames is simplified to iterating over the capture object.\n\n```python\nimport cv2\nimport cvpack\n\nwith cvpack.VideoCapture("video.mp4") as cap:\n    with cvpack.VideoWriter("reversed.mp4", fourcc=int(cap.fourcc), fps=cap.fps) as writer:\n        for frame in cap:\n            flipped = cv2.flip(frame, 0)\n            writer.write(flipped)\n```\n',
    'author': 'Alexander Reynolds',
    'author_email': 'alex@theory.shop',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/alkasm/cvpack',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
